;;; inf-lucy.el --- run the Lucid Synchrone toplevel in an Emacs buffer

;; Xavier Leroy, july 1993.

;; modified by Jacques Garrigue, july 1997.

(require 'comint)
(require 'lucy)

;; User modifiable variables

;; Whether you want the output buffer to be diplayed when you send a phrase

(defvar lucy-display-when-eval t
  "*If true, display the inferior lucy buffer when evaluating expressions.")


;; End of User modifiable variables


(defvar inferior-lucy-mode-map nil)
(if inferior-lucy-mode-map nil
  (setq inferior-lucy-mode-map
        (copy-keymap comint-mode-map)))

;; Augment Lucy mode, so you can process Lucy code in the source files.

(defvar inferior-lucy-program "lucy"
  "*Program name for invoking an inferior Lucy from Emacs.")

(defun inferior-lucy-mode ()
  "Major mode for interacting with an inferior Lucy process.
Runs a Lucy toplevel as a subprocess of Emacs, with I/O through an
Emacs buffer. A history of input phrases is maintained. Phrases can
be sent from another buffer in Lucy mode.

\\{inferior-lucy-mode-map}"
  (interactive)
  (comint-mode)
  (setq comint-prompt-regexp "^# ?")
  (setq major-mode 'inferior-lucy-mode)
  (setq mode-name "Inferior Lucy")
  (make-local-variable 'paragraph-start)
  (setq paragraph-start (concat "^$\\|" page-delimiter))
  (make-local-variable 'paragraph-separate)
  (setq paragraph-separate paragraph-start)
  (make-local-variable 'paragraph-ignore-fill-prefix)
  (setq paragraph-ignore-fill-prefix t)
  (make-local-variable 'require-final-newline)
  (setq require-final-newline t)
  (make-local-variable 'comment-start)
  (setq comment-start "(*")
  (make-local-variable 'comment-end)
  (setq comment-end "*)")
  (make-local-variable 'comment-column)
  (setq comment-column 40)
  (make-local-variable 'comment-start-skip)
  (setq comment-start-skip "(\\*+ *")
  (make-local-variable 'parse-sexp-ignore-comments)
  (setq parse-sexp-ignore-comments nil)
  (use-local-map inferior-lucy-mode-map)
  (run-hooks 'inferior-lucy-mode-hooks))


(defconst inferior-lucy-buffer-subname "inferior-lucy")
(defconst inferior-lucy-buffer-name
  (concat "*" inferior-lucy-buffer-subname "*"))

;; for compatibility with xemacs 

(defun lucy-sit-for (second &optional mili redisplay)
   (if (and (boundp 'running-xemacs) running-xemacs)
       (sit-for (if mili (+ second (* mili 0.001)) second) redisplay)
     (sit-for second mili redisplay)))

;; To show result of evaluation at toplevel

(defvar inferior-lucy-output nil)
(defun inferior-lucy-signal-output (s)
  (if (string-match "[^ ]" s) (setq inferior-lucy-output t)))

(defun inferior-lucy-mode-output-hook ()
  (setq comint-output-filter-functions
        (list (function inferior-lucy-signal-output))))
(add-hook 'inferior-lucy-mode-hooks 'inferior-lucy-mode-output-hook)

;; To launch lucy whenever needed

(defun lucy-run-process-if-needed (&optional cmd)
  (if (comint-check-proc inferior-lucy-buffer-name) nil
    (if (not cmd)
        (if (comint-check-proc inferior-lucy-buffer-name)
            (setq cmd inferior-lucy-program)
          (setq cmd (read-from-minibuffer "Lucy toplevel to run: "
                                          inferior-lucy-program))))
    (setq inferior-lucy-program cmd)
    (let ((cmdlist (inferior-lucy-args-to-list cmd))
          (process-connection-type nil))
      (set-buffer (apply (function make-comint)
                         inferior-lucy-buffer-subname
                         (car cmdlist) nil (cdr cmdlist)))
      (inferior-lucy-mode)
      (display-buffer inferior-lucy-buffer-name)
      t)
    (setq lucy-shell-active t)
    ))

;; patched to from original run-lucy sharing code with
;;  lucy-run-process-when-needed

(defun run-lucy (&optional cmd)
  "Run an inferior Lucy process.
Input and output via buffer `*inferior-lucy*'."
  (interactive
   (list (if (not (comint-check-proc inferior-lucy-buffer-name))
             (read-from-minibuffer "Lucy toplevel to run: "
                                   inferior-lucy-program))))
  (lucy-run-process-if-needed cmd)
  (switch-to-buffer-other-window inferior-lucy-buffer-name))


(defun inferior-lucy-args-to-list (string)
  (let ((where (string-match "[ \t]" string)))
    (cond ((null where) (list string))
          ((not (= where 0))
           (cons (substring string 0 where)
                 (inferior-lucy-args-to-list (substring string (+ 1 where)
                                                        (length string)))))
          (t (let ((pos (string-match "[^ \t]" string)))
               (if (null pos)
                   nil
                 (inferior-lucy-args-to-list (substring string pos
                                                        (length string)))))))))

(defun inferior-lucy-show-subshell ()
  (interactive)
  (lucy-run-process-if-needed)
  (display-buffer inferior-lucy-buffer-name)
  ; Added by Didier to move the point of inferior-lucy to end of buffer
  (let ((buf (current-buffer))
        (lucy-buf  (get-buffer inferior-lucy-buffer-name))
        (count 0))
    (while
        (and (< count 10)
             (not (equal (buffer-name (current-buffer))
                         inferior-lucy-buffer-name)))
      (next-multiframe-window)
      (setq count (+ count 1)))
    (if  (equal (buffer-name (current-buffer))
                inferior-lucy-buffer-name)
        (end-of-buffer))
    (while
        (> count 0)
      (previous-multiframe-window)
      (setq count (- count 1)))
    )
)

;; patched by Didier to move cursor after evaluation 

(defun inferior-lucy-eval-region (start end)
  "Send the current region to the inferior Lucy process."
  (interactive "r")
  (save-excursion (lucy-run-process-if-needed))
  (save-excursion
    (goto-char end)
    (lucy-skip-comments-backward)
    (comint-send-region inferior-lucy-buffer-name start (point))
    ;; normally, ";;" are part of the region
    (if (and (>= (point) 2)
             (prog2 (backward-char 2) (looking-at ";;")))
        (comint-send-string inferior-lucy-buffer-name "\n")
      (comint-send-string inferior-lucy-buffer-name ";;\n"))
    ;; the user may not want to see the output buffer
    (if lucy-display-when-eval
        (display-buffer inferior-lucy-buffer-name t))))

;; jump to errors produced by lucy compiler

(defun inferior-lucy-goto-error (start end)
  "Jump to the location of the last error as indicated by inferior toplevel."
  (interactive "r")
  (let ((loc (+ start
                (save-excursion
                  (set-buffer (get-buffer inferior-lucy-buffer-name))
                  (re-search-backward
                   (concat comint-prompt-regexp
                           "[ \t]*Characters[ \t]+\\([0-9]+\\)-[0-9]+:$"))
                  (string-to-int (match-string 1))))))
    (goto-char loc)))


;;; orgininal inf-lucy.el ended here

;; as eval-phrase, but ignores errors.

(defun inferior-lucy-just-eval-phrase (arg &optional min max)
  "Send the phrase containing the point to the LUCY process.
With prefix-arg send as many phrases as its numeric value,
ignoring possible errors during evaluation.

Optional arguments min max defines a region within which the phrase
should lies."
  (interactive "p")
  (let ((beg))
    (while (> arg 0)
      (setq arg (- arg 1))
      (setq beg  (lucy-find-phrase min max))
      (lucy-eval-region beg (point)))
    beg))

(defvar lucy-previous-output nil
  "tells the beginning of output in the shell-output buffer, so that the
output can be retreived later, asynchronously.")

;; enriched version of eval-phrase, to repport errors.

(defun inferior-lucy-eval-phrase (arg &optional min max)
  "Send the phrase containing the point to the LUCY process.
With prefix-arg send as many phrases as its numeric value, 
If an error occurs during evalutaion, stop at this phrase and
repport the error. 

Return nil if noerror and position of error if any.

If arg's numeric value is zero or negative, evaluate the current phrase
or as many as prefix arg, ignoring evaluation errors. 
This allows to jump other erroneous phrases. 

Optional arguments min max defines a region within which the phrase
should lies."
  (interactive "p")
  (if (save-excursion (lucy-run-process-if-needed))
      (progn
        (setq inferior-lucy-output nil)
        (lucy-wait-output 10 1)))
  (if (< arg 1) (inferior-lucy-just-eval-phrase (max 1 (- 0 arg)) min max)
    (let ((proc (get-buffer-process inferior-lucy-buffer-name))
          (buf (current-buffer))
          previous-output orig beg end err)
      (save-window-excursion
        (while (and (> arg 0) (not err))
          (setq previous-output (marker-position (process-mark proc)))
          (setq lucy-previous-output previous-output)
          (setq inferior-lucy-output nil)
          (setq orig (inferior-lucy-just-eval-phrase 1 min max))
          (lucy-wait-output)
          (switch-to-buffer inferior-lucy-buffer-name  nil)
          (goto-char previous-output)
          (cond ((re-search-forward
                  " *Characters \\([01-9][01-9]*\\)-\\([1-9][01-9]*\\):\n[^W]"
                  (point-max) t)
                 (setq beg (string-to-int (lucy-match-string 1)))
                 (setq end (string-to-int (lucy-match-string 2)))
                 (switch-to-buffer buf)
                 (goto-char orig)
                 (forward-byte end)
                 (setq end (point))
                 (goto-char orig)
                 (forward-byte beg)
                 (setq beg (point))
                 (setq err beg)
                 )
                ((looking-at
                  "Toplevel input:\n[>]\\([^\n]*\\)\n[>]\\(\\( *\\)^*\\)\n")
                 (let ((expr (lucy-match-string 1))
                       (column (-   (match-end 3) (match-beginning 3)))
                       (width (-   (match-end 2) (match-end 3))))
                   (if (string-match  "^\\(.*\\)[<]EOF[>]$" expr)
                       (setq expr (substring expr (match-beginning 1) (match-end 1))))
                   (switch-to-buffer buf)
                   (re-search-backward
                    (concat "^" (regexp-quote expr) "$")
                    (- orig 10))
                   (goto-char (+ (match-beginning 0) column))
                   (setq end (+ (point) width)))
                 (setq err beg))
                ((looking-at
                  "Toplevel input:\n>[.]*\\([^.].*\n\\)\\([>].*\n\\)*[>]\\(.*[^.]\\)[.]*\n")
                 (let* ((e1 (lucy-match-string 1))
                        (e2 (lucy-match-string 3))
                        (expr
                         (concat
                          (regexp-quote e1) "\\(.*\n\\)*" (regexp-quote e2))))
                   (switch-to-buffer buf)
                   (re-search-backward expr orig 'move)
                   (setq end (match-end 0)))
                 (setq err beg))
                (t
                 (switch-to-buffer buf)))
          (setq arg (- arg 1))
          )
        (pop-to-buffer inferior-lucy-buffer-name)
        (if err
            (goto-char (point-max))
          (goto-char previous-output)
          (goto-char (point-max)))
        (pop-to-buffer buf))
      (if err (progn (beep) (lucy-overlay-region (point) end))
        (if inferior-lucy-output
            (message "No error")
          (message "No output yet...")
          ))
      err)))

(defun lucy-overlay-region (beg end &optional wait)
  (interactive "%r")
  (cond ((fboundp 'make-overlay)
         (if lucy-error-overlay ()
           (setq lucy-error-overlay (make-overlay 1 1))
           (overlay-put lucy-error-overlay 'face 'region))
         (unwind-protect
             (progn
               (move-overlay lucy-error-overlay beg end (current-buffer))
               (beep) (if wait (read-event) (lucy-sit-for 60)))
           (delete-overlay lucy-error-overlay)))))  

;; wait some amount for ouput, that is, until inferior-lucy-output is set
;; to true. Hence, interleaves sitting for shorts delays and checking the
;; flag. Give up after some time. Typing into the source buffer will cancel 
;; waiting, i.e. may report 'No result yet' 

(defun lucy-wait-output (&optional before after)
  (let ((c 1))
    (lucy-sit-for 0 (or before 1))
    (let ((c 1))
      (while (and (not inferior-lucy-output) (< c 99) (lucy-sit-for 0 c t))
        (setq c (+ c 1))))
    (lucy-sit-for (or after 0) 1)))

;; To insert the last output from lucy at point
(defun lucy-insert-last-output ()
  "Insert the result of the evaluation of previous phrase"
  (interactive)
  (let ((pos (process-mark (get-buffer-process inferior-lucy-buffer-name))))
  (insert-buffer-substring inferior-lucy-buffer-name
                           lucy-previous-output (- pos 2))))

;; additional bindings
  
;(let ((map (lookup-key lucy-mode-map [menu-bar lucy])))
;  (define-key map [indent-buffer] '("Indent buffer" . lucy-indent-buffer))
;  (define-key map [eval-buffer] '("Eval buffer" . lucy-eval-buffer))
;) 
;(define-key lucy-mode-map "\C-c\C-b" 'lucy-eval-buffer)


(provide 'inf-lucy)
