;;; Run lucydebug under Emacs
;;; Derived from gdb.el.
;;; gdb.el is Copyright (C) 1988 Free Software Foundation, Inc, and is part
;;; of GNU Emacs
;;; Modified by Jerome Vouillon, 1994.
;;; Modified by Ian T. Zimmerman, 1996.
;;; Modified by Xavier Leroy, 1997.

;; This file is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 1, or (at your option)
;; any later version.

;; This file is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;;itz 04-06-96 I pondered basing this on gud. The potential advantages
;;were: automatic bugfix , keymaps and menus propagation.
;;Disadvantages: gud is not so clean itself, there is little common
;;functionality it abstracts (most of the stuff is done in the
;;debugger specific parts anyway), and, most seriously, gud sees it
;;fit to add C-x C-a bindings to the _global_ map, so there would be a
;;conflict between lucydebug and gdb, for instance. While it's OK to
;;assume that a sane person doesn't use gdb and dbx at the same time,
;;it's not so OK (IMHO) for gdb and lucydebug.

;; Xavier Leroy, 21/02/97: adaptation to lucydebug.

(require 'comint)
(require 'shell)
(require 'lucy)
(require 'derived)
(require 'thingatpt)

;;; Variables.

(defvar lucydebug-last-frame)
(defvar lucydebug-delete-prompt-marker)
(defvar lucydebug-filter-accumulator nil)
(defvar lucydebug-last-frame-displayed-p)
(defvar lucydebug-filter-function)

(defvar lucydebug-prompt-pattern "^(ocd) *"
  "A regexp to recognize the prompt for lucydebug.")

(defvar lucydebug-overlay-event nil
  "Overlay for displaying the current event.")
(defvar lucydebug-overlay-under nil
  "Overlay for displaying the current event.")
(defvar lucydebug-event-marker nil
  "Marker for displaying the current event.")

(defvar lucydebug-track-frame t
  "*If non-nil, always display current frame position in another window.")

(cond
 (window-system
  (make-face 'lucydebug-event)
  (make-face 'lucydebug-underline)
  (if (not (face-differs-from-default-p 'lucydebug-event))
      (invert-face 'lucydebug-event))
  (if (not (face-differs-from-default-p 'lucydebug-underline))
      (set-face-underline-p 'lucydebug-underline t))
  (setq lucydebug-overlay-event (make-overlay 1 1))
  (overlay-put lucydebug-overlay-event 'face 'lucydebug-event)
  (setq lucydebug-overlay-under (make-overlay 1 1))
  (overlay-put lucydebug-overlay-under 'face 'lucydebug-underline))
 (t
  (setq lucydebug-event-marker (make-marker))
  (setq overlay-arrow-string "=>")))

;;; Lucydebug mode.

(define-derived-mode lucydebug-mode comint-mode "Inferior CDB"

  "Major mode for interacting with an inferior Lucydebug process.

The following commands are available:

\\{lucydebug-mode-map}

\\[lucydebug-display-frame] displays in the other window
the last line referred to in the lucydebug buffer.

\\[lucydebug-step], \\[lucydebug-back] and \\[lucydebug-next], in the lucydebug window,
call lucydebug to step, backstep or next and then update the other window
with the current file and position.

If you are in a source file, you may select a point to break
at, by doing \\[lucydebug-break].

Commands:
Many commands are inherited from comint mode.
Additionally we have:

\\[lucydebug-display-frame] display frames file in other window
\\[lucydebug-step] advance one line in program
C-x SPACE sets break point at current line."

  (mapcar 'make-local-variable
          '(lucydebug-last-frame-displayed-p  lucydebug-last-frame
            lucydebug-delete-prompt-marker lucydebug-filter-function
           lucydebug-filter-accumulator paragraph-start))
  (setq
   lucydebug-last-frame nil
   lucydebug-delete-prompt-marker (make-marker)
   lucydebug-filter-accumulator ""
   lucydebug-filter-function 'lucydebug-marker-filter
   comint-prompt-regexp lucydebug-prompt-pattern
   comint-dynamic-complete-functions (cons 'lucydebug-complete
                                           comint-dynamic-complete-functions)
   paragraph-start comint-prompt-regexp
   lucydebug-last-frame-displayed-p t)
  (make-local-variable 'shell-dirtrackp)
  (setq shell-dirtrackp t)
  (setq comint-input-sentinel 'shell-directory-tracker))

;;; Keymaps.

(defun lucydebug-numeric-arg (arg)
  (and arg (prefix-numeric-value arg)))

(defmacro def-lucydebug (name key &optional doc args)

  "Define lucydebug-NAME to be a command sending NAME ARGS and bound
to KEY, with optional doc string DOC.  Certain %-escapes in ARGS are
interpreted specially if present.  These are:

  %m    module name of current module.
  %d    directory of current source file.
  %c    number of current character position
  %e    text of the lucy variable surrounding point.

  The `current' source file is the file of the current buffer (if
we're in a lucy buffer) or the source file current at the last break
or step (if we're in the lucydebug buffer), and the `current' module
name is the filename stripped of any *.ls* suffixes (this assumes the
usual correspondence between module and file naming is observed).  The
`current' position is that of the current buffer (if we're in a source
file) or the position of the last break or step (if we're in the
lucydebug buffer).

If a numeric is present, it overrides any ARGS flags and its string
representation is simply concatenated with the COMMAND."

  (let* ((fun (intern (format "lucydebug-%s" name))))
    (list 'progn
          (if doc
              (list 'defun fun '(arg)
                    doc
                    '(interactive "P")
                    (list 'lucydebug-call name args
                          '(lucydebug-numeric-arg arg))))
          (list 'define-key 'lucydebug-mode-map
                (concat "\C-c" key)
                (list 'quote fun))
          (list 'define-key 'lucy-mode-map
                (concat "\C-x\C-a" key)
                (list 'quote fun)))))

(def-lucydebug "step"   "\C-s"  "Step one event forward.")
(def-lucydebug "backstep" "\C-k" "Step one event backward.")
(def-lucydebug "run"    "\C-r"  "Run the program.")
(def-lucydebug "reverse" "\C-v" "Run the program in reverse.")
(def-lucydebug "last"   "\C-l"  "Go to latest time in execution history.")
(def-lucydebug "backtrace" "\C-t" "Print the call stack.")
(def-lucydebug "finish" "\C-f"  "Finish executing current function.")
(def-lucydebug "print"  "\C-p"  "Print value of symbol at point."       "%e")
(def-lucydebug "display" "\C-d" "Display value of symbol at point."     "%e")
(def-lucydebug "next"   "\C-n"  "Step one event forward (skip functions)")
(def-lucydebug "up"     "<"  "Go up N stack frames (numeric arg) with display")
(def-lucydebug "down"  ">" "Go down N stack frames (numeric arg) with display")
(def-lucydebug "break"  "\C-b"  "Set breakpoint at current line."
  "@ \"%m\" # %c")

(defun lucydebug-mouse-display (click)
  "Display value of $NNN clicked on."
  (interactive "e")
  (let* ((start (event-start click))
         (window (car start))
         (pos (car (cdr start)))
         symb)
    (save-excursion
      (select-window window)
      (goto-char pos)
      (setq symb (thing-at-point 'symbol))
      (if (string-match "^\\$[0-9]+$" symb)
          (lucydebug-call "display" symb)))))

(define-key lucydebug-mode-map [mouse-2] 'lucydebug-mouse-display)

(defun lucydebug-kill-filter (string)
  ;gob up stupid questions :-)
  (setq lucydebug-filter-accumulator
        (concat lucydebug-filter-accumulator string))
  (if (not (string-match "\\(.* \\)(y or n) "
                         lucydebug-filter-accumulator)) nil
    (setq lucydebug-kill-output
          (cons t (match-string 1 lucydebug-filter-accumulator)))
    (setq lucydebug-filter-accumulator ""))
  (if (string-match comint-prompt-regexp lucydebug-filter-accumulator)
      (let ((output (substring lucydebug-filter-accumulator
                               (match-beginning 0))))
        (setq lucydebug-kill-output
              (cons nil (substring lucydebug-filter-accumulator 0
                                   (1- (match-beginning 0)))))
        (setq lucydebug-filter-accumulator "")
        output)
    ""))

(def-lucydebug "kill"   "\C-k")

(defun lucydebug-kill ()
  "Kill the program."
  (interactive)
  (let ((lucydebug-kill-output))
    (save-excursion
      (set-buffer current-lucydebug-buffer)
      (let ((proc (get-buffer-process (current-buffer)))
            (lucydebug-filter-function 'lucydebug-kill-filter))
        (lucydebug-call "kill")
        (while (not (and lucydebug-kill-output
                         (zerop (length lucydebug-filter-accumulator))))
          (accept-process-output proc))))
    (if (not (car lucydebug-kill-output))
        (error (cdr lucydebug-kill-output))
      (sit-for 0 300)
      (lucydebug-call-1 (if (y-or-n-p (cdr lucydebug-kill-output)) "y" "n")))))
;;FIXME: lucydebug doesn't output the Hide marker on kill

(defun lucydebug-goto-filter (string)
  ;accumulate onto previous output
  (setq lucydebug-filter-accumulator
        (concat lucydebug-filter-accumulator string))
  (if (not (string-match (concat "\\(\n\\|\\`\\)[ \t]*\\([0-9]+\\)[ \t]+"
                          lucydebug-goto-position
                          "[ \t]*\\(before\\|after\\)\n")
                  lucydebug-filter-accumulator)) nil
    (setq lucydebug-goto-output
          (match-string 2 lucydebug-filter-accumulator))
    (setq lucydebug-filter-accumulator
          (substring lucydebug-filter-accumulator (1- (match-end 0)))))
  (if (not (string-match comint-prompt-regexp
                         lucydebug-filter-accumulator)) nil
    (setq lucydebug-goto-output (or lucydebug-goto-output 'fail))
    (setq lucydebug-filter-accumulator ""))
  (if (string-match "\n\\(.*\\)\\'" lucydebug-filter-accumulator)
      (setq lucydebug-filter-accumulator
            (match-string 1 lucydebug-filter-accumulator)))
  "")

(def-lucydebug "goto" "\C-g")
(defun lucydebug-goto (&optional time)

  "Go to the execution time TIME.

Without TIME, the command behaves as follows: In the lucydebug buffer,
if the point at buffer end, goto time 0\; otherwise, try to obtain the
time from context around point. In a lucy mode buffer, try to find the
time associated in execution history with the current point location.

With a negative TIME, move that many lines backward in the lucydebug
buffer, then try to obtain the time from context around point."

  (interactive "P")
  (cond
   (time
    (let ((ntime (lucydebug-numeric-arg time)))
      (if (>= ntime 0) (lucydebug-call "goto" nil ntime)
        (save-selected-window
          (select-window (get-buffer-window current-lucydebug-buffer))
          (save-excursion
            (if (re-search-backward "^Time : [0-9]+ - pc : [0-9]+ "
                                    nil t (- 1 ntime))
                (lucydebug-goto nil)
              (error "I don't have %d times in my history"
                     (- 1 ntime))))))))
   ((eq (current-buffer) current-lucydebug-buffer)
      (let ((time (cond
                   ((eobp) 0)
                   ((save-excursion
                      (beginning-of-line 1)
                      (looking-at "^Time : \\([0-9]+\\) - pc : [0-9]+ "))
                    (string-to-int (match-string 1)))
                   ((string-to-int (lucydebug-format-command "%e"))))))
        (lucydebug-call "goto" nil time)))
   (t
    (let ((module (lucydebug-module-name (buffer-file-name)))
          (lucydebug-goto-position (int-to-string (1- (point))))
          (lucydebug-goto-output) (address))
      ;get a list of all events in the current module
      (save-excursion
        (set-buffer current-lucydebug-buffer)
        (let* ((proc (get-buffer-process (current-buffer)))
               (lucydebug-filter-function 'lucydebug-goto-filter))
          (lucydebug-call-1 (concat "info events " module))
          (while (not (and lucydebug-goto-output
                      (zerop (length lucydebug-filter-accumulator))))
            (accept-process-output proc))
          (setq address (if (eq lucydebug-goto-output 'fail) nil
                          (re-search-backward
                           (concat "^Time : \\([0-9]+\\) - pc : "
                                   lucydebug-goto-output
                                   " - module "
                                   module "$") nil t)
                          (match-string 1)))))
      (if address (lucydebug-call "goto" nil (string-to-int address))
        (error "No time at %s at %s" module lucydebug-goto-position))))))


(defun lucydebug-delete-filter (string)
  (setq lucydebug-filter-accumulator
        (concat lucydebug-filter-accumulator string))
  (if (not (string-match
            (concat "\\(\n\\|\\`\\)[ \t]*\\([0-9]+\\)[ \t]+[0-9]+[ \t]*in "
                    (regexp-quote lucydebug-delete-file)
                    ", character "
                    lucydebug-delete-position "\n")
                  lucydebug-filter-accumulator)) nil
    (setq lucydebug-delete-output
          (match-string 2 lucydebug-filter-accumulator))
    (setq lucydebug-filter-accumulator
          (substring lucydebug-filter-accumulator (1- (match-end 0)))))
  (if (not (string-match comint-prompt-regexp
                         lucydebug-filter-accumulator)) nil
    (setq lucydebug-delete-output (or lucydebug-delete-output 'fail))
    (setq lucydebug-filter-accumulator ""))
  (if (string-match "\n\\(.*\\)\\'" lucydebug-filter-accumulator)
      (setq lucydebug-filter-accumulator
            (match-string 1 lucydebug-filter-accumulator)))
  "")


(def-lucydebug "delete" "\C-d")

(defun lucydebug-delete (&optional arg)
  "Delete the breakpoint numbered ARG.

Without ARG, the command behaves as follows: In the lucydebug buffer,
try to obtain the time from context around point. In a lucy mode
buffer, try to find the breakpoint associated with the current point
location.

With a negative ARG, look for the -ARGth breakpoint pattern in the
lucydebug buffer, then try to obtain the breakpoint info from context
around point."

  (interactive "P")
  (cond
   (arg
    (let ((narg (lucydebug-numeric-arg arg)))
      (if (> narg 0) (lucydebug-call "delete" nil narg)
        (save-excursion
          (set-buffer current-lucydebug-buffer)
          (if (re-search-backward "^Breakpoint [0-9]+ at [0-9]+ : file "
                                  nil t (- 1 narg))
              (lucydebug-delete nil)
            (error "I don't have %d breakpoints in my history"
                     (- 1 narg)))))))
   ((eq (current-buffer) current-lucydebug-buffer)
    (let* ((bpline "^Breakpoint \\([0-9]+\\) at [0-9]+ : file ")
           (arg (cond
                 ((eobp)
                  (save-excursion (re-search-backward bpline nil t))
                  (string-to-int (match-string 1)))
                 ((save-excursion
                    (beginning-of-line 1)
                    (looking-at bpline))
                  (string-to-int (match-string 1)))
                 ((string-to-int (lucydebug-format-command "%e"))))))
      (lucydebug-call "delete" nil arg)))
   (t
    (let ((lucydebug-delete-file
           (concat (lucydebug-format-command "%m") ".ls"))
          (lucydebug-delete-position (lucydebug-format-command "%c")))
      (save-excursion
        (set-buffer current-lucydebug-buffer)
        (let ((proc (get-buffer-process (current-buffer)))
              (lucydebug-filter-function 'lucydebug-delete-filter)
              (lucydebug-delete-output))
          (lucydebug-call-1 "info break")
          (while (not (and lucydebug-delete-output
                           (zerop (length
                                   lucydebug-filter-accumulator))))
            (accept-process-output proc))
          (if (eq lucydebug-delete-output 'fail)
              (error "No breakpoint in %s at %s"
                     lucydebug-delete-file
                     lucydebug-delete-position)
            (lucydebug-call "delete" nil
                            (string-to-int lucydebug-delete-output)))))))))

(defun lucydebug-complete-filter (string)
  (setq lucydebug-filter-accumulator
        (concat lucydebug-filter-accumulator string))
  (while (string-match "\\(\n\\|\\`\\)\\(.+\\)\n"
                       lucydebug-filter-accumulator)
    (setq lucydebug-complete-list
          (cons (match-string 2 lucydebug-filter-accumulator)
                lucydebug-complete-list))
    (setq lucydebug-filter-accumulator
          (substring lucydebug-filter-accumulator
                     (1- (match-end 0)))))
  (if (not (string-match comint-prompt-regexp
                         lucydebug-filter-accumulator)) nil
    (setq lucydebug-complete-list
          (or lucydebug-complete-list 'fail))
    (setq lucydebug-filter-accumulator ""))
  (if (string-match "\n\\(.*\\)\\'" lucydebug-filter-accumulator)
      (setq lucydebug-filter-accumulator
            (match-string 1 lucydebug-filter-accumulator)))
  "")

(defun lucydebug-complete ()

  "Perform completion on the lucydebug command preceding point."

  (interactive)
  (let* ((end (point))
         (command (save-excursion
                    (beginning-of-line)
                    (and (looking-at comint-prompt-regexp)
                         (goto-char (match-end 0)))
                    (buffer-substring (point) end)))
         (lucydebug-complete-list nil) (command-word))

    ;; Find the word break.  This match will always succeed.
    (string-match "\\(\\`\\| \\)\\([^ ]*\\)\\'" command)
    (setq command-word (match-string 2 command))

    ;itz 04-21-96 if we are trying to complete a word of nonzero
    ;length, chop off the last character. This is a nasty hack, but it
    ;works - in general, not just for this set of words: the comint
    ;call below will weed out false matches - and it avoids further
    ;mucking with lucydebug's lexer.
    (if (> (length command-word) 0)
        (setq command (substring command 0 (1- (length command)))))

    (let ((lucydebug-filter-function 'lucydebug-complete-filter))
      (lucydebug-call-1 (concat "complete " command))
      (set-marker lucydebug-delete-prompt-marker nil)
      (while (not (and lucydebug-complete-list
                       (zerop (length lucydebug-filter-accumulator))))
        (accept-process-output (get-buffer-process
                                (current-buffer)))))
    (if (eq lucydebug-complete-list 'fail)
        (setq lucydebug-complete-list nil))
    (setq lucydebug-complete-list
          (sort lucydebug-complete-list 'string-lessp))
    (comint-dynamic-simple-complete command-word lucydebug-complete-list)))

(define-key lucydebug-mode-map "\C-l" 'lucydebug-refresh)
(define-key lucydebug-mode-map "\t" 'comint-dynamic-complete)
(define-key lucydebug-mode-map "\M-?" 'comint-dynamic-list-completions)

(define-key lucy-mode-map "\C-x " 'lucydebug-break)


(defvar current-lucydebug-buffer nil)


;;;###autoload
(defvar lucydebug-command-name "lucydebug"
  "Pathname for executing lucydebug.")

;;;###autoload
(defun lucydebug (path)
  "Run lucydebug on program FILE in buffer *lucydebug-FILE*.
The directory containing FILE becomes the initial working directory
and source-file directory for lucydebug.  If you wish to change this, use
the lucydebug commands `cd DIR' and `directory'."
  (interactive "fRun lucydebug on file: ")
  (setq path (expand-file-name path))
  (let ((file (file-name-nondirectory path)))
    (pop-to-buffer (concat "*lucydebug-" file "*"))
    (setq default-directory (file-name-directory path))
    (message "Current directory is %s" default-directory)
    (make-comint (concat "lucydebug-" file)
                 (substitute-in-file-name lucydebug-command-name)
                 nil
                 "-emacs" "-cd" default-directory file)
    (set-process-filter (get-buffer-process (current-buffer))
                        'lucydebug-filter)
    (set-process-sentinel (get-buffer-process (current-buffer))
                          'lucydebug-sentinel)
    (lucydebug-mode)
    (lucydebug-set-buffer)))

(defun lucydebug-set-buffer ()
  (if (eq major-mode 'lucydebug-mode)
      (setq current-lucydebug-buffer (current-buffer))
    (save-selected-window (pop-to-buffer current-lucydebug-buffer))))

;;; Filter and sentinel.

(defun lucydebug-marker-filter (string)
  (setq lucydebug-filter-accumulator
        (concat lucydebug-filter-accumulator string))
  (let ((output "") (begin))
    ;; Process all the complete markers in this chunk.
    (while (setq begin
                 (string-match
                  "\032\032\\(H\\|M\\(.+\\):\\(.+\\):\\(before\\|after\\)\\)\n"
                  lucydebug-filter-accumulator))
      (setq lucydebug-last-frame
            (if (char-equal ?H (aref lucydebug-filter-accumulator
                                     (1+ (1+ begin)))) nil
              (list (match-string 2 lucydebug-filter-accumulator)
                    (string-to-int
                     (match-string 3 lucydebug-filter-accumulator))
                    (string= "before"
                             (match-string 4
                                           lucydebug-filter-accumulator))))
            output (concat output
                           (substring lucydebug-filter-accumulator
                                      0 begin))
            ;; Set the accumulator to the remaining text.
            lucydebug-filter-accumulator (substring
                                          lucydebug-filter-accumulator
                                          (match-end 0))
            lucydebug-last-frame-displayed-p nil))

    ;; Does the remaining text look like it might end with the
    ;; beginning of another marker?  If it does, then keep it in
    ;; lucydebug-filter-accumulator until we receive the rest of it.  Since we
    ;; know the full marker regexp above failed, it's pretty simple to
    ;; test for marker starts.
    (if (string-match "\032.*\\'" lucydebug-filter-accumulator)
        (progn
          ;; Everything before the potential marker start can be output.
          (setq output (concat output (substring lucydebug-filter-accumulator
                                                 0 (match-beginning 0))))

          ;; Everything after, we save, to combine with later input.
          (setq lucydebug-filter-accumulator
                (substring lucydebug-filter-accumulator (match-beginning 0))))

      (setq output (concat output lucydebug-filter-accumulator)
            lucydebug-filter-accumulator ""))

    output))

(defun lucydebug-filter (proc string)
  (let ((output))
    (if (buffer-name (process-buffer proc))
        (let ((process-window))
          (save-excursion
            (set-buffer (process-buffer proc))
            ;; If we have been so requested, delete the debugger prompt.
            (if (marker-buffer lucydebug-delete-prompt-marker)
                (progn
                  (delete-region (process-mark proc)
                                 lucydebug-delete-prompt-marker)
                  (set-marker lucydebug-delete-prompt-marker nil)))
            (setq output (funcall lucydebug-filter-function string))
            ;; Don't display the specified file unless
            ;; (1) point is at or after the position where output appears
            ;; and (2) this buffer is on the screen.
            (setq process-window (and lucydebug-track-frame
                                      (not lucydebug-last-frame-displayed-p)
                                      (>= (point) (process-mark proc))
                                      (get-buffer-window (current-buffer))))
            ;; Insert the text, moving the process-marker.
            (comint-output-filter proc output))
          (if process-window
              (save-selected-window
                (select-window process-window)
                (lucydebug-display-frame)))))))

(defun lucydebug-sentinel (proc msg)
  (cond ((null (buffer-name (process-buffer proc)))
         ;; buffer killed
         ;; Stop displaying an arrow in a source file.
         (lucydebug-remove-current-event)
         (set-process-buffer proc nil))
        ((memq (process-status proc) '(signal exit))
         ;; Stop displaying an arrow in a source file.
         (lucydebug-remove-current-event)
         ;; Fix the mode line.
         (setq mode-line-process
               (concat ": "
                       (symbol-name (process-status proc))))
         (let* ((obuf (current-buffer)))
           ;; save-excursion isn't the right thing if
           ;;  process-buffer is current-buffer
           (unwind-protect
               (progn
                 ;; Write something in *compilation* and hack its mode line,
                 (set-buffer (process-buffer proc))
                 ;; Force mode line redisplay soon
                 (set-buffer-modified-p (buffer-modified-p))
                 (if (eobp)
                     (insert ?\n mode-name " " msg)
                   (save-excursion
                     (goto-char (point-max))
                     (insert ?\n mode-name " " msg)))
                 ;; If buffer and mode line will show that the process
                 ;; is dead, we can delete it now.  Otherwise it
                 ;; will stay around until M-x list-processes.
                 (delete-process proc))
             ;; Restore old buffer, but don't restore old point
             ;; if obuf is the cdb buffer.
             (set-buffer obuf))))))


(defun lucydebug-refresh (&optional arg)
  "Fix up a possibly garbled display, and redraw the mark."
  (interactive "P")
  (lucydebug-display-frame)
  (recenter arg))

(defun lucydebug-display-frame ()
  "Find, obey and delete the last filename-and-line marker from CDB.
The marker looks like \\032\\032FILENAME:CHARACTER\\n.
Obeying it means displaying in another window the specified file and line."
  (interactive)
  (lucydebug-set-buffer)
  (if (not lucydebug-last-frame)
      (lucydebug-remove-current-event)
    (lucydebug-display-line (car lucydebug-last-frame)
                            (car (cdr lucydebug-last-frame))
                            (car (cdr (cdr lucydebug-last-frame)))))
  (setq lucydebug-last-frame-displayed-p t))

;; Make sure the file named TRUE-FILE is in a buffer that appears on the screen
;; and that its character CHARACTER is visible.
;; Put the mark on this character in that buffer.

(defun lucydebug-display-line (true-file character kind)
  (let* ((pre-display-buffer-function nil) ; screw it, put it all in one screen
         (pop-up-windows t)
         (buffer (find-file-noselect true-file))
         (window (display-buffer buffer t))
         (pos))
    (save-excursion
      (set-buffer buffer)
      (save-restriction
        (widen)
        (setq pos (+ (point-min) character))
        (lucydebug-set-current-event pos (current-buffer) kind))
      (cond ((or (< pos (point-min)) (> pos (point-max)))
             (widen)
             (goto-char pos))))
    (set-window-point window pos)))

;;; Events.

(defun lucydebug-remove-current-event ()
  (if window-system
      (progn
        (delete-overlay lucydebug-overlay-event)
        (delete-overlay lucydebug-overlay-under))
    (setq overlay-arrow-position nil)))

(defun lucydebug-set-current-event (pos buffer before)
  (if window-system
      (if before
          (progn
            (move-overlay lucydebug-overlay-event pos (1+ pos) buffer)
            (move-overlay lucydebug-overlay-under
                          (+ pos 1) (+ pos 3) buffer))
        (move-overlay lucydebug-overlay-event (1- pos) pos buffer)
        (move-overlay lucydebug-overlay-under (- pos 3) (- pos 1) buffer))
    (save-excursion
      (set-buffer buffer)
      (goto-char pos)
      (beginning-of-line)
      (move-marker lucydebug-event-marker (point))
      (setq overlay-arrow-position lucydebug-event-marker))))

;;; Miscellaneous.

(defun lucydebug-module-name (filename)
  (substring filename (string-match "\\([^/]*\\)\\.ls$" filename) (match-end 1)))

;;; The lucydebug-call function must do the right thing whether its
;;; invoking keystroke is from the lucydebug buffer itself (via
;;; major-mode binding) or a lucy buffer.  In the former case, we want
;;; to supply data from lucydebug-last-frame.  Here's how we do it:

(defun lucydebug-format-command (str)
  (let* ((insource (not (eq (current-buffer) current-lucydebug-buffer)))
        (frame (if insource nil lucydebug-last-frame)) (result))
    (while (and str (string-match "\\([^%]*\\)%\\([mdcep]\\)" str))
      (let ((key (string-to-char (substring str (match-beginning 2))))
            (cmd (substring str (match-beginning 1) (match-end 1)))
            (subst))
        (setq str (substring str (match-end 2)))
        (cond
         ((eq key ?m)
          (setq subst (lucydebug-module-name
                       (if insource (buffer-file-name) (nth 0 frame)))))
         ((eq key ?d)
          (setq subst (file-name-directory
                       (if insource (buffer-file-name) (nth 0 frame)))))
         ((eq key ?c)
          (setq subst (int-to-string
                       (if insource (1- (point)) (nth 1 frame)))))
         ((eq key ?e)
          (setq subst (thing-at-point 'symbol))))
        (setq result (concat result cmd subst))))
    ;; There might be text left in STR when the loop ends.
    (concat result str)))

(defun lucydebug-call (command &optional fmt arg)
  "Invoke lucydebug COMMAND displaying source in other window.

Certain %-escapes in FMT are interpreted specially if present.
These are:

  %m    module name of current module.
  %d    directory of current source file.
  %c    number of current character position
  %e    text of the lucy variable surrounding point.

  The `current' source file is the file of the current buffer (if
we're in a lucy buffer) or the source file current at the last break
or step (if we're in the lucydebug buffer), and the `current' module
name is the filename stripped of any *.ls* suffixes (this assumes the
usual correspondence between module and file naming is observed).  The
`current' position is that of the current buffer (if we're in a source
file) or the position of the last break or step (if we're in the
lucydebug buffer).

If ARG is present, it overrides any FMT flags and its string
representation is simply concatenated with the COMMAND."

  ;; Make sure debugger buffer is displayed in a window.
  (lucydebug-set-buffer)
  (message "Command: %s" (lucydebug-call-1 command fmt arg)))

(defun lucydebug-call-1 (command &optional fmt arg)

  ;; Record info on the last prompt in the buffer and its position.
  (save-excursion
    (set-buffer current-lucydebug-buffer)
    (goto-char (process-mark (get-buffer-process current-lucydebug-buffer)))
    (let ((pt (point)))
      (beginning-of-line)
      (if (looking-at comint-prompt-regexp)
          (set-marker lucydebug-delete-prompt-marker (point)))))
  (let ((cmd (cond
              (arg (concat command " " (int-to-string arg)))
              (fmt (lucydebug-format-command
                    (concat command " " fmt)))
              (command))))
    (process-send-string (get-buffer-process current-lucydebug-buffer)
                         (concat cmd "\n"))
    cmd))


(provide 'lucydebug)
