(* a PID controler for the inverted pendulum *)
(* $Id: controller.ls,v 1.1 2007/01/04 17:55:07 pouzet Exp $ *)

open Graphics
open Data
open Pendulum
open Chrono

(*                       --------       select      theta0                *)
(*                       |       |        |           |                   *)
(*   manual (x0, y0) ----| force |--------\           |                   *)
(*           p           |       |        |      ------------             *)
(*                       ---------        |      |          |             *)
(*                                        |------| Pendulum |----         *)
(*                          -------       |      |          |   |         *)
(*                          |     |       /      ------------   |         *)
(* desired_theta = 0 -- +/--| PID | ----->                      |         *)
(*                      /\  |     |                             |         *)
(*                       |  -------                             |         *)
(*                       ---------------------------------------|         *)
(*                                                                        *)

(* force *)
let node force x0 y0 = (d2x0, d2y0) where
  rec
    d2x0 = deriv (deriv x0)
  and
    d2y0 = deriv (deriv y0)
 
(* conversion *)
let conv theta = (x, y) where
  rec x = l *. sin theta
  and y = l *. cos theta

(* PID law *)
let node pid dt kp ki kd delta = d2x where
  rec
    d2x = delta *. kp +. aux *. ki +. daux *. kd
  and
    aux = delta +. (0.0 -> pre aux)
  and
    daux = 0.0 -> (delta -. pre delta) /. dt

let static kp = 0.0001
let static ki = 1.0
let static kd = 0.0
let static dt = 0.01

(* switch according to [toggle] *)
let node switch toggle = o where
  automaton
      True -> do o = true unless toggle then False
    | False -> do o = false unless toggle then True
  end

(*
(* the first version is purely data-flow *)
let node controller1 toggle theta0 ix0 iy0 =
  let rec
      md2x0, md2y0 = force x0 y0
  and
      error_x = l *. sin (0.0 -> 0.0 -. pre theta)
  and
      f_x = pid dt kp ki kd error_x
  and
      x0, y0, id2x0, id2y0 = 
         if switch toggle then x0, y0, x_x, 0.0 else x0, y0, md2x0, md2y0
  and
      theta = pendulum theta0 id2x0 id2y0 in
  
  let x = x0 +. l*.(sin theta) in
  let y = y0 +. l*.(cos theta) in
    (x0, y0, x, y, theta)
*)

(* the second one mixes data-flow and automata *)
(* the first version is purely data-flow *)
let node controller toggle theta0 ix0 iy0 =
  let rec 
      automaton
          Manual -> 
	    do
	      x0 = ix0
	    and
	      y0 = iy0
	    (* and
	      d2x0, d2y0 = force x0 y0*)
	    until false then Auto
	| Auto ->
	    let d_x = pid dt kp ki kd error_x in
	    (*let x0' = integr (integr cd2x0) in*)
	    do
	      (* d2x0 = cd2x0
	    and
	      d2y0 = 0.0
	    and *)
	      x0 = (last x0) +. d_x
	    and
	      y0 = (last y0)
	    until ((*print_string "b";flush stdout;*)toggle) then Manual
      end
  and
      d2x0, d2y0 = force x0 y0
  and
      theta = mod_float (pendulum theta0 d2x0 d2y0) (2.0 *. 3.14156)
  and
      error_x = l *. sin (0.0 -> 0.0 -. pre theta) in
  
  let x = x0 +. l*.(sin theta) in
  let y = y0 +. l*.(cos theta) in
    print_float theta; print_string " "; flush stdout;
    (ix0, iy0, x0, y0, x, y, theta)

(* outputs for sim2chro *)
let node chronogram (static oc) toggle (mx0, my0, x0, y0, x, y, theta) =
  let initialisation, running = Chrono.sim2chro oc in
  let static program = "Pendulm" in
  let static def_inputs = 
    [ "m_x0", Tfloat;
      "m_y0", Tfloat;
      "click", Tbool ] in
  let static def_outputs =
    [ "x0", Tfloat;
      "y0", Tfloat;
      "x", Tfloat;
      "y", Tfloat;
      "theta", Tfloat] in
  let inputs =
    [ Vfloat(mx0); Vfloat(my0); Vbool(toggle) ] in
  let outputs =
    [ Vfloat(x0); Vfloat(y0); Vfloat(x); Vfloat(y); Vfloat(theta) ] in
    
  match true -> false with
      true -> initialisation program def_inputs def_outputs
    | false -> ()
  end;
  running (inputs, outputs)

(* the main entry *)
(* the system switch from the automatic to manual mode (and conversely) *)
(* on a mouse click *)
let node automatic oc status =
  let toggle = status.button in
  match toggle with 
      true -> print_string "aaa\n";flush stdout 
    | false -> () end;
  let x0 = float status.mouse_x /. 10.0 in
  let y0 = float status.mouse_y /. 10.0 in
  let (mx0, my0, x0, y0, x, y, theta) = controller toggle 0.0 x0 y0 in
  chronogram oc toggle (mx0, my0, x0, y0, x, y, theta);
  let p = make_pend x0 y0 x y in
  draw_pendulum p (p fby p)
